-- PLBARS 1..10 reserved, [1] sw width, [2] sw height, [3] sw class height, [4] show level, [5] RGBA substrate, [6] update freq;
-- PLBARS[7] - show actionbar
-- PLBARS 11..20 - player1, 21..30 - player2, 31..40 - player3, 41..50 - player4, 51..60 - healthbar, 61 manabar, 71 actionbar,
-- 81 currbar, 91 - extrabar, 101 - inv bar
-- for each entry [10*i+1] left, [10*i+2] bottom; health/mana/cast bars: [10*i+3] width, [10*i+4] height, [10*i+5] color, 
-- curr bar, extra bar, etc [10*i+6] show / hide
-- /console taintlog 1 

local ver = "0.6"

local helpText = {
"|cFFffff00PlayerBars short help and command list:|r",
"|cFFffff00For full documentation please visit|r",
"|cFF00ffffhttp://www.curse.com/addons/wow/playerbars|r",
"Use Shift + Mouse Left Click to drag elements",
"|cFF00ffff/pbars|r - toggles all elements",
"|cFF00ffff/pbars hw <number>|r - sets health bar width",
"|cFF00ffff/pbars hh <number>|r - sets health bar height",
"|cFF00ffff/pbars hc <number> <number> <number> <number>|r",
"- sets health bar rgba color",
"- use |cFF00ffffmw|r, |cFF00ffffmh|r, |cFF00ffffmc|r instead of |cFF00ffffhw|r, |cFF00ffffhh|r, |cFF00ffffhc|r to change power bar",
"- use |cFF00ffffaw|r, |cFF00ffffah|r, |cFF00ffffac|r instead of |cFF00ffffhw|r, |cFF00ffffhh|r, |cFF00ffffhc|r to change cast bar",
"|cFF00ffff/pbars c <number> <number> <number> <number>|r",
"- sets health/mana/cast bar background rgba color",
"|cFFffff00Cast bar:|r",
"|cFF00ffff/pbars a|r - switches between default and custom cast bar",
"|cFFffff00Currency bar:|r",
"|cFF00ffff/pbars cb|r - toggles currency bar",
"|cFFffff00Inventory bar:|r",
"|cFF00ffff/pbars i|r - toggles inventory bar",
"|cFF00ffff/pbars iw <number>|r - sets bar width",
"|cFF00ffff/pbars ih <number>|r - sets bar height",
"|cFFffff00Group commands:|r",
"|cFF00ffff/pbars g|r - toggles group swatches",
"|cFF00ffff/pbars gw <number>|r - sets swatch width",
"|cFF00ffff/pbars gh <number>|r - sets swatch height",
"|cFFffff00Block commands:|r",
"|cFF00ffff/pbars b|r - justifies health/mana/cast bars and puts them together",
"|cFF00ffff/pbars bgh <number>|r - lines up swatches in a row separated by value",
"|cFF00ffff/pbars bgv <number>|r - lines up swatches in a column separated by value"
}

local battleStatus = false
local timer = 0.0
local timer2 = 0.0
local ratio = {1, 1, 1, 1,  1, 1, 0}
local rleft, rbottom, rwidth, rheight
local currBar = {5, 1, 20} -- margin, gap, subbar height
local invBar = {5, 1, 20} -- margin, gap, title height
local extBar  = {3} -- margin
local imgMar = 3 -- icon margin

local playerName = UnitName("player")

-- stores ind, slotID, slotTexture, itemLink, human name of the slot
local slots = {
	{"HeadSlot",0,"","","Head"}, -- left side
	{"NeckSlot",0,"","","Neck"},
	{"ShoulderSlot",0,"","","Shoulders"},
	{"BackSlot",0,"","","Back"},
	{"ChestSlot",0,"","","Chest"},
	{"ShirtSlot",0,"","","Shirt"},
	{"TabardSlot",0,"","","Tabard"},
	{"WristSlot",0,"","","Wrists"},
	{"MainHandSlot",0,"","","Main hand"},
	{"HandsSlot",0,"","","Hands"}, -- right side
	{"WaistSlot",0,"","","Waist"},
	{"LegsSlot",0,"","","Legs"},
	{"FeetSlot",0,"","","Feet"},
	{"Finger0Slot",0,"","","Middle finger"},
	{"Finger1Slot",0,"","","Pinky"}, 
	{"Trinket0Slot",0,"","","Left earhole"},
	{"Trinket1Slot",0,"","","Right earhole"},
	{"SecondaryHandSlot",0,"","","Off hand"}
	}

local curList = {823, 824, 1101}

local icon = {
	[1] = "Interface\\Icons\\inv_apexis_draenor",
	[2] = "Interface\\Icons\\inv_garrison_resource",
	[3] = "Interface\\Icons\\garrison_oil",
	[4] = "Interface\\Icons\\inv_misc_coin_01",
	[5] = "Interface\\Buttons\\UI-RefreshButton"
}

local xRGBA = {
	["high"] =  {0, 1, 0, 0.5},
	["mid"] =   {1, 1, 0, 0.5},
	["low"] =   {1, 0.6, 0, 0.5},
	["crit"] =  {1, 0, 0, 0.5},
	["zero"] =  {0, 0, 0, 0.5},
	["off"] =   {0.6, 0.6, 0.6, 0.3},
	["dark"] =  {0, 0, 0, 0.5},
	["pitch"] = {0, 0, 0, 0.66},
	["black"] = {0, 0, 0, 1},
	-- text colors
	["yell"] =  {1, 1, 0, 0.75},
	["yell_dim"] = {1, 1, 0, 0.5},
	["white"] = {1, 1, 1, 0.66},
	["white_dim"] = {1, 1, 1, 0.15}
}

-- default bar colors
local barColor = {
	["HEALTH"] = {0,0.7,0,0.8},
	["MANA"] = {0,0,0.7,0.8},
	["RAGE"] = {0.7,0,0,0.8},
	["FOCUS"] = {0.6,0.4,0.3,0.8},
	["ENERGY"] = {0.7,0.7,0,0.8},
	["COMBO_POINTS"] = {0.5,0.5,0.5,0.8},
	["RUNES"] = {0.5,0.5,0.5,0.8},
	["RUNIC_POWER"] = {0.5,0.5,0.5,0.8},
	["SOUL_SHARDS"] = {0.5,0.5,0.5,0.8},
	["ECLIPSE"] = {0.5,0.5,0.5,0.8},
	["HOLY_POWER"] = {0.5,0.5,0.5,0.8},
	["AMMOSLOT"] = {1,0,0,0.8},
	["FUEL"] = {1,1,0,0.8},
	["CASTBAR"] = {1,0,1,0.8},
	["SUBSTR"] = {1,1,1,0.25},
	["EXTRA"] = {1,1,0,0.66}
}
local classPT = {"RAGE", "MANA", "FOCUS", "ENERGY", "MANA", "RUNIC_POWER", "MANA", "MANA", "MANA", "ENERGY", "MANA"}

-- 1..4 swatches, frame ref, text ref, class frame ref, party index
-- 5 healthbar, 6 manabar, 7 actionbar, 8 main cur bar, 81 - 84 sub bars, 9 extra bar, 10 inv bar, 1001 - 1018 sub bars
local barNum = 10
xRef = {}
for i = 1, barNum do
	xRef[i] = {}
end

local className = {"WR", "PL", "HT", "RG", "PR", "DK", "SH", "MG", "LK", "MN", "DR"}  
local roleName = {["DAMAGER"] = "D", ["HEALER"] = "H", ["NONE"] = "X", ["TANK"] = "T"}

local function GetGroupInfo()
    local unit, classInd, role, name, hRatio
	for i = 1, 4 do
		unit = "party" .. i
		if UnitExists(unit) then
			name = UnitName(unit)
			if name ~= playerName then
				-- DAMAGER, HEALER, NONE, TANK
				role = UnitGroupRolesAssigned(unit)
				_, _, classInd = UnitClass(unit)
				hRatio = UnitHealth(name) / UnitHealthMax(name)
				
				if not UnitIsConnected(name) or UnitHealth(name) < 2 then hRatio = 0 end
				
				if hRatio > 0.75 then
					xRef[i][2]:SetTexture(unpack(xRGBA["high"]))
				elseif hRatio > 0.5 then
					xRef[i][2]:SetTexture(unpack(xRGBA["mid"]))
				elseif hRatio > 0.25 then
					xRef[i][2]:SetTexture(unpack(xRGBA["low"]))
				elseif hRatio > 0 then
					xRef[i][2]:SetTexture(unpack(xRGBA["crit"]))
				else
					xRef[i][2]:SetTexture(unpack(xRGBA["zero"]))
				end
				
				xRef[i][1].text:SetTextColor(unpack(xRGBA["white"]))
				xRef[i][1].text:SetText(roleName[role])
				if classInd ~= nill then
					xRef[i][3].text:SetText(className[classInd] .. "-" .. UnitLevel(unit))
				else
					xRef[i][3].text:SetText("??" .. "-" .. UnitLevel(unit))
				end
			end
		else
			xRef[i][2]:SetTexture(unpack(xRGBA["zero"]))
			xRef[i][1].text:SetTextColor(unpack(xRGBA["white_dim"]))
			xRef[i][1].text:SetText("-")
			xRef[i][3].text:SetText("-")
		end
	end
end

local function resetFrame(frRef, frPoint, frPar, parPoint, xx, yy)
		local left, bottom, _, height = frRef:GetBoundsRect()
		local x = xx or left
		local y = yy or bottom + height
		if frPar == nil then frPar = WorldFrame end
		if frPoint == nil then frPoint = "TOPLEFT" end
		if parPoint == nil then parPoint = "BOTTOMLEFT" end
		frRef:ClearAllPoints()
		frRef:SetPoint(frPoint, frPar, parPoint, x, y)
end

local function safeSetWidth(frRef, w)
	if w < 1 then w = 1 end
	frRef:SetWidth(w)
end

local function toggleInvBar(force)
	local switch = force or PLBARS[10*10 + 6]
	if switch == 1 then
		xRef[10][1]:Show()
		for k = 0, 18 do
			xRef[1000+k][1]:Show()
		end
	else
		xRef[10][1]:Hide()
		for k = 0, 18 do
			xRef[1000+k][1]:Hide()
		end
	end
end

local function setInvBarWHY()
	local slotWidth = (PLBARS[10*10 + 3] - 3*invBar[1])/2
	local slotHeight = (PLBARS[10*10 + 4] - 2*invBar[1] - 8*invBar[2] - invBar[3]-invBar[2])/9
	local slotY = PLBARS[10*10 + 4] - invBar[1] - invBar[3] - invBar[2]
	return slotWidth, slotHeight, slotY
end

local function setInvBar(type)
	local slotWidth, slotHeight, slotY = setInvBarWHY()
	local itemRow = 1
	if type == "w" then
		-- title
		xRef[1000][1]:SetPoint("TOPLEFT", xRef[10][1], "BOTTOMLEFT", invBar[1], PLBARS[10*10 + 4] - invBar[1])
		xRef[1000][1]:SetWidth(PLBARS[10*10 + 3] - 2*invBar[1])
		for k = 1, 18 do
			xRef[1000+k][1]:SetWidth(slotWidth)
			if k > 9 then
				xRef[1000+k][1]:SetPoint("TOPLEFT", xRef[10][1], "BOTTOMLEFT", invBar[1] + itemRow*(slotWidth+invBar[1]), slotY)
				slotY = slotY - slotHeight - invBar[2]
			end
			-- resize text frame 
			xRef[1000+k][5]:SetWidth(slotWidth - slotHeight)
		end
	elseif type == "h" then
		-- title
		xRef[1000][1]:SetPoint("TOPLEFT", xRef[10][1], "BOTTOMLEFT", invBar[1], PLBARS[10*10 + 4] - invBar[1])
		xRef[1000][1]:SetHeight(invBar[3])
		for k = 1, 18 do
			xRef[1000+k][1]:SetHeight(slotHeight)
			itemRow = 0
			if k > 9 then itemRow = 1 end
			if k == 10 then slotY = select(3,setInvBarWHY()) end
			xRef[1000+k][1]:SetPoint("TOPLEFT", xRef[10][1], "BOTTOMLEFT", invBar[1] + itemRow*(slotWidth+invBar[1]), slotY)
			slotY = slotY - slotHeight - invBar[2]
			-- resize icons
			xRef[1000+k][6]:SetSize(slotHeight-imgMar*2, slotHeight-imgMar*2) 
			-- resize text frame 
			xRef[1000+k][5]:SetSize(slotWidth - slotHeight,slotHeight)
			xRef[1000+k][5]:SetPoint("BOTTOMLEFT", xRef[1000+k][1], "BOTTOMLEFT", slotHeight, 0)
		end
	end
end

local iLevelPattern = "^Item Level (%d+)"
local scanGT -- scanning tooltip
local function checkInvItem(slotID, slotTex, slotLink)
	local iTxt, iTex, iLink, iColor, line
	scanGT = scanGT or CreateFrame("GameTooltip", "ScanTooltip", nil, "GameTooltipTemplate")
	scanGT:SetOwner(UIParent, "ANCHOR_NONE")
	iLink = GetInventoryItemLink("player", slotID)
	if iLink == nil then
		iColor = xRGBA["pitch"]
		iTex = slotTex
		iTxt = "none"
	else
		iColor = {("0x"..strsub(iLink, 5, 6))/1024, ("0x"..strsub(iLink, 7, 8))/1024, ("0x"..strsub(iLink, 9, 10))/1024, 0.8}
		iTex = GetInventoryItemTexture("player", slotID)
		iTxt = "-"
		-- get item level from tooltip
		scanGT:SetHyperlink(iLink)
		for i = 2, scanGT:NumLines() do
			line = _G["ScanTooltipTextLeft"..i]:GetText()
			iTxt = strmatch(line, iLevelPattern)
			if iTxt and iTxt ~= "" then
				break
			end
		end
	end
	return iTxt, iTex, iColor
end

local function updateInvBar(slotID)
	local iTxt, iTex, iColor
	if slotID ~= nil then
		for k, v in pairs(slots) do
			if v[2] == slotID then
				iTxt, iTex, iColor = checkInvItem(v[2], v[3], v[4]) -- slotID, slotTex, stored iLink
				xRef[1000+k][2]:SetTexture(unpack(iColor))
				xRef[1000+k][5].text:SetText(" "..iTxt)
				xRef[1000+k][7]:SetTexture(iTex)
				break
			end
		end
	else
		for k, v in pairs(slots) do
			iTxt, iTex, iColor = checkInvItem(v[2], v[3], v[4]) -- slotID, slotTex, stored iLink
			xRef[1000+k][2]:SetTexture(unpack(iColor))
			xRef[1000+k][5].text:SetText(" "..iTxt)
			xRef[1000+k][7]:SetTexture(iTex)
		end
	end
	-- item level
	local iOver, iEquip = GetAverageItemLevel()
	local tText = "-"
	if iOver ~= nil and iEquip ~= nil then tText = format("%.1f/%.1f",iEquip,iOver) end
	xRef[1000][5].text:SetText(tText)
end

local function setActionBar()
	if PLBARS[7] == 1 then
		xRef[7][1]:Show()
		_G["CastingBarFrame"].showCastbar = false
	else
		xRef[7][1]:Hide()
		_G["CastingBarFrame"].showCastbar = true
	end
end

local function setExtraBar()
	if PLBARS[10*9+6] == 1 then
		if not battleStatus then
			xRef[9][1]:Show()
			xRef[9][5]:Show()
			ExtraActionBarFrame.button:Hide()
		end
		-- design the bar
		local EABF = ExtraActionBarFrame
		local tex = EABF.button.icon:GetTexture()
		if not HasExtraActionBar() then 
			xRef[9][3]:SetTexture(0,0,0,0)
			xRef[9][2]:SetTexture(unpack(PLBARS[10*9 + 5]))
		else
			if tex ~= nil then xRef[9][3]:SetTexture(tex) end
			xRef[9][2]:SetTexture(unpack(barColor["EXTRA"]))
		end
	else
		if not battleStatus then
			xRef[9][1]:Hide()
			xRef[9][5]:Hide()
			ExtraActionBarFrame.button:Show()
		end
	end
end



local function onUpdate(self,elapsed)

	-- curr bar update once a sec
	timer2 = timer2 + elapsed
	if xRef[8][1]:IsVisible() and timer2 > 1.0 then
		local currVal, money, gold, silver, copper
		for i = 1, 3 do
			currVal = select(2, GetCurrencyInfo(curList[i]))
			xRef[80 + i][5].text:SetText(tostring(currVal))
		end
		money = GetMoney()
		gold = floor(abs(money / 10000))
		silver = floor(abs(mod(money / 100, 100)))
		copper = floor(abs(mod(money, 100)))
		xRef[80 + 4][5].text:SetText(format("%d %d %d", gold, silver, copper))
		timer2 = 0
	end

	timer = timer + elapsed

	if timer > PLBARS[6] then
		ratio[5] = UnitHealth(playerName) / UnitHealthMax(playerName)
		ratio[6] = UnitPower(playerName) / UnitPowerMax(playerName)
		for i = 5, 6 do
			safeSetWidth(xRef[i][3], PLBARS[10*i + 3] * ratio[i])
		end

		-- castbar
		local CBname, CBtext, startTime, endTime
		ratio[7] = 0
		CBname, _, CBtext, _, startTime, endTime = UnitCastingInfo("player")
		if CBname ~= nil then
			ratio[7] = (GetTime()*1000 - startTime) / (endTime - startTime)
		else
			CBname, _, CBtext, _, startTime, endTime = UnitChannelInfo("player")
			if CBname ~= nil then
				ratio[7] = (GetTime()*1000 - startTime) / (endTime - startTime)
				CBtext = CBname
			end
		end
		if CBtext ~= nil then
			xRef[7][5].text:SetText(" " .. strupper(CBtext))
		else
			xRef[7][5].text:SetText(" ")
		end
		safeSetWidth(xRef[7][3], ratio[7] * PLBARS[10*7 + 3])
		
		-- group state
		if xRef[1][1]:IsVisible() then
			if IsInGroup() then 
				GetGroupInfo()
			else
				-- reset empty values
				for i = 1, 4 do
					xRef[i][2]:SetTexture(unpack(xRGBA["off"]))
					xRef[i][1].text:SetTextColor(unpack(xRGBA["white_dim"]))
					xRef[i][1].text:SetText("G" .. i)
					xRef[i][3].text:SetText(" ")
				end
			end
		end
		
		timer = 0
	end
end


local function SetGameTooltipDefaultAnchor(tooltip, tpoint, parent, point, x, y)      
  tooltip:SetOwner(parent, "ANCHOR_NONE")
  tooltip:SetPoint(tpoint, UIParent, point, x, y)
  tooltip.default = 1
end

local function onMouseLeave(self, motion)
	GameTooltip:Hide()
end

local function onMouseEnter(self, motion)
	if self == xRef[9][4] and HasExtraActionBar() then
		local btnID = ActionButton_GetPagedID(ExtraActionBarFrame.button)
		local splID = GetActionInfo(btnID) -- or 180455
		if splID ~= nil then
			GameTooltip_SetDefaultAnchor(GameTooltip, UIParent)
			GameTooltip:SetSpellByID(splID)
			GameTooltip:Show()
		end
	elseif self ~= xRef[9][4] then
		GameTooltip:Hide()
		local par = self:GetParent()
		local parInd = tonumber(strsub(par:GetName(), 12))
		local left, bottom, width, height = xRef[10][1]:GetBoundsRect()
		if left+width < GetScreenWidth()/2 then
			SetGameTooltipDefaultAnchor(GameTooltip, "TOPLEFT", UIParent, "BOTTOMLEFT", left+width, bottom + height)
		elseif left > GetScreenWidth()/2 then 
			SetGameTooltipDefaultAnchor(GameTooltip, "BOTTOMRIGHT", UIParent, "BOTTOMLEFT", left, bottom)
		end
		if GameTooltip:SetInventoryItem("player", slots[parInd][2]) == nil then 
			GameTooltip:SetText(slots[parInd][5].." (empty)", 1, 1, 1, 0.5)
		end
		GameTooltip:Show()
	end
end

local function onMouseDown(self, button)
	if button == "LeftButton" and IsShiftKeyDown() then
		if self == xRef[9][4] then self = xRef[9][1] end -- move extrabar when clicked on sec button
		if not self.isMoving then
			self:StartMoving()
			self.isMoving = true
		end
	end
end

local function onMouseUp(self, button)
	if button == "LeftButton" then
		-- update inv bar
		if self == xRef[1000][6] then 
			updateInvBar()
			return
		end
		if self == xRef[9][4] then self = xRef[9][1] end
		if self.isMoving then
			self:StopMovingOrSizing()
			self.isMoving = false
			-- memo x,y; 4 swatches left, bottom(incl frClass h); 5,6,7 - health/mana/cast; 8-curr, 9-extra, 10 inv
			for i = 1, barNum do
				PLBARS[10 * i + 1], PLBARS[10 * i + 2] = xRef[i][1]:GetBoundsRect()
			end
		end
	end
end

local function setFrameProps(frRef)
	frRef:SetMovable(true)
	frRef:EnableMouse(true)
	frRef:SetScript("OnMouseDown", onMouseDown)
	frRef:SetScript("OnMouseUp", onMouseUp)
end

local function createSecureButton(bname, x, y, w, h, RGBA, img, parent, btnNum, passive)
	local secBtn, secBtnTex
	local parent = parent or WorldFrame
	
	secBtn = CreateFrame("Button", bname, parent, "SecureActionButtonTemplate")
	secBtn:SetAttribute("type", "action")
	secBtn:SetAttribute("action", btnNum)
	secBtn:SetSize(w, h) 
	secBtn:SetPoint("BOTTOMLEFT", parent, "BOTTOMLEFT", x, y)
	secBtnTex = secBtn:CreateTexture() 
	secBtnTex:SetAllPoints()
	if img ~= nil then 
		secBtnTex:SetTexture(img)
	else
		secBtnTex:SetTexture(unpack(RGBA))
	end
	secBtn.background = secBtnTex
	
	if passive == nil then setFrameProps(secBtn) end
	
	return secBtn, secBtnTex
end

local function createFrame(fname, x, y, w, h, RGBA, txt, overlay, r, img, parent, passive)
	local frSub, frSubTex,  frOver, frOverTex,  frTxt,  frIco, frIcoTex
	local txOff = 0
	local parent = parent or WorldFrame
	
	frSub = CreateFrame("Frame", fname, UIParent)
	frSub:SetClampedToScreen(true)
	frSub:SetSize(w, h) 
	frSub:SetPoint("TOPLEFT", parent, "BOTTOMLEFT", x, y)
	frSubTex = frSub:CreateTexture()
	frSubTex:SetAllPoints()
	if overlay == nil then
		frSubTex:SetTexture(unpack(RGBA))
	else
		frSubTex:SetTexture(unpack(PLBARS[5]))
	end
	frSub.background = frSubTex

	-- icon frame
	if img ~= nil then
		frIco = CreateFrame("Frame", "Icon", frSub)
		frIco:SetClampedToScreen(true)
		frIco:SetSize(h-imgMar*2, h-imgMar*2) 
		frIco:SetPoint("BOTTOMLEFT", frSub, "BOTTOMLEFT", imgMar, imgMar)
		frIcoTex = frIco:CreateTexture()
		frIcoTex:SetAllPoints()
		frIcoTex:SetTexture(img)
		frIco.background = frIcoTex
		txOff = h
	end

	if overlay ~= nil then
		frOver = CreateFrame("Frame", "SliderFrame", frSub)
		frOver:SetClampedToScreen(true)
		safeSetWidth(frOver, w*r)
		frOver:SetHeight(h) 
		frOver:SetPoint("BOTTOMLEFT", frSub, "BOTTOMLEFT", 0, 0)
		frOverTex = frOver:CreateTexture()
		frOverTex:SetAllPoints()
		frOverTex:SetTexture(unpack(RGBA))
		frOver.background = frOverTex
	end
	
	if txt ~= nil then
		frTxt = CreateFrame("Frame", "TextFrame", frSub)
		frTxt:SetClampedToScreen(true)
		frTxt:SetSize(w-txOff, h) 
		frTxt:SetPoint("BOTTOMLEFT", frSub, "BOTTOMLEFT", txOff, 0)

		frTxt.text = frTxt:CreateFontString(nil,"ARTWORK")
		frTxt.text:SetFontObject("Tooltip_Small")
		frTxt.text:SetTextColor(unpack(xRGBA["yell"]))
		frTxt.text:SetMaxLines(1)
		frTxt.text:SetJustifyH("LEFT")
		frTxt.text:SetText(txt)
		frTxt.text:SetAllPoints()
	end
	
	if passive == nil then setFrameProps(frSub) end
	
	return frSub, frSubTex,  frOver, frOverTex,  frTxt,  frIco, frIcoTex
end

local function createSwatch(fname, x, y, w, h, txt, txt2, RGBA)
	
	local frRef, texRef = createFrame(fname, x, y, w, h, RGBA)
	local frClass, frClassTex

	frClass = CreateFrame("Frame", "ClassFrame", frRef)
	frClass:SetSize(w, PLBARS[3]) 
	frClass:SetPoint("BOTTOMLEFT", frRef, "BOTTOMLEFT", 0, -PLBARS[3])
	frClassTex = frClass:CreateTexture()
	frClassTex:SetAllPoints()
	frClassTex:SetTexture(unpack(xRGBA["dark"])) 
	frClass.background = frClassTex
	
	if txt ~= nil then
		frRef.text = frRef:CreateFontString(nil,"ARTWORK")
		frRef.text:SetFontObject("GameFontHighlightLarge")
		frRef.text:SetTextColor(unpack(xRGBA["white_dim"]))
		frRef.text:SetText(txt)
		frRef.text:SetAllPoints()
	end
	
	if txt2 ~= nil then
		frClass.text = frClass:CreateFontString(nil,"ARTWORK")
		frClass.text:SetFontObject("Tooltip_Small")
		frClass.text:SetTextColor(unpack(xRGBA["yell"]))
		frClass.text:SetText(txt2)
		frClass.text:SetAllPoints()
	end
	
	setFrameProps(frRef)
	
	return frRef, texRef, frClass
end

local waitingList = {}
local eframe = CreateFrame("Frame", "EventFrame", UIParent)
eframe:RegisterEvent("ADDON_LOADED")
eframe:RegisterEvent("UPDATE_EXTRA_ACTIONBAR")
eframe:RegisterEvent("PLAYER_REGEN_DISABLED")
eframe:RegisterEvent("PLAYER_REGEN_ENABLED")
eframe:RegisterEvent("PLAYER_ENTERING_WORLD")
eframe:RegisterEvent("ITEM_UPGRADE_MASTER_UPDATE")
eframe:RegisterEvent("PLAYER_EQUIPMENT_CHANGED")
eframe:RegisterEvent("GET_ITEM_INFO_RECEIVED")
function eframe:OnEvent(event, arg)

--print(event, arg)

if event == "GET_ITEM_INFO_RECEIVED" then
	waitingList[arg] = nil
	local iReady = true
	for k,v in pairs(waitingList) do
		if v ~= nil then
			iReady = false
			break
		end
	end
	if iReady then updateInvBar() end
end

if event == "PLAYER_ENTERING_WORLD" then
	-- inv setup
	local iLink, itemID
	local allLoaded = true
	for k,v in pairs(slots) do
		v[2], v[3] = GetInventorySlotInfo(v[1]) -- slot id, texture
		itemID = GetInventoryItemID("player", v[2])
		if itemID == nil then -- empty slot
			iLink = nil
		else
			_, iLink = GetItemInfo(itemID)
			if iLink == nil then -- wait for the server
				waitingList[itemID] = 1
				allLoaded = false
			end
		end
		v[4] = iLink
	end
	if allLoaded then updateInvBar() end
	return
end

if event == "PLAYER_REGEN_DISABLED" then 
	battleStatus = true
	return
end
if event == "PLAYER_REGEN_ENABLED" then 
	battleStatus = false
	return
end

if event == "PLAYER_EQUIPMENT_CHANGED" then
	updateInvBar(arg)
	return
end

if event == "ITEM_UPGRADE_MASTER_UPDATE" and arg == "player" then
	updateInvBar()
	return
end

if event == "UPDATE_EXTRA_ACTIONBAR" then
	if xRef[9][1] ~= nil then
		setExtraBar()
	end
	return
end

if event == "ADDON_LOADED" and arg == "PlayerBars" then

	if PLBARS == nil then PLBARS = {} end
	
	-- width, height
	PLBARS[10*5 + 3] = PLBARS[10*5 + 3] or 265
	PLBARS[10*5 + 4] = PLBARS[10*5 + 4] or 12
	PLBARS[10*6 + 3] = PLBARS[10*6 + 3] or 265
	PLBARS[10*6 + 4] = PLBARS[10*6 + 4] or 6
	PLBARS[10*7 + 3] = PLBARS[10*7 + 3] or 265
	PLBARS[10*7 + 4] = PLBARS[10*7 + 4] or 10
	
	PLBARS[10*8 + 3] = PLBARS[10*8 + 3] or 80
	-- height of 4 subbars + top/bot margin + gaps
	PLBARS[10*8 + 4] = currBar[3]*4 + currBar[1]*2 + (4-1)*currBar[2] 
	
	PLBARS[10*9 + 3] = PLBARS[10*9 + 3] or 32
	PLBARS[10*9 + 4] = PLBARS[10*9 + 4] or PLBARS[10*9 + 3]
	
	-- inv
	PLBARS[10*10 + 3] = PLBARS[10*10 + 3] or 120
	PLBARS[10*10 + 4] = PLBARS[10*10 + 4] or 220
	
	-- color
	local manaType = classPT[select(3, UnitClass("player"))]
	if manaType == nil then manaType = "MANA" end
	
	PLBARS[10*6 + 5] = PLBARS[10*6 + 5] or barColor[manaType]
	PLBARS[10*5 + 5] = PLBARS[10*5 + 5] or barColor["HEALTH"]
	PLBARS[10*7 + 5] = PLBARS[10*7 + 5] or barColor["CASTBAR"]
	PLBARS[10*8 + 5] = PLBARS[10*8 + 5] or barColor["SUBSTR"]
	PLBARS[10*9 + 5] = PLBARS[10*9 + 5] or xRGBA["off"]
	PLBARS[10*10 + 5] = PLBARS[10*10 + 5] or barColor["SUBSTR"]
	
	-- freq
	PLBARS[6] = PLBARS[6] or 0.1
	-- all substrates
	PLBARS[5] = PLBARS[5] or barColor["SUBSTR"]
	-- swatches
	PLBARS[1] = PLBARS[1] or 40
	PLBARS[2] = PLBARS[2] or 20
	PLBARS[3] = PLBARS[3] or 12
	PLBARS[4] = PLBARS[4] or true
	-- show actionbar
	PLBARS[7] = PLBARS[7] or 1
	-- show currbar, extra bar, inv bar
	for i = 8, barNum do
		PLBARS[10*i + 6] = PLBARS[10*i + 6] or 1
	end
	
	local centerX = GetScreenWidth()/2
	local centerY = GetScreenHeight()/2
	
	local startXX = {}
	local startXY = {}
	for i = 1, 4 do
		startXX[i] = PLBARS[10*i + 1] or centerX + 85 + (PLBARS[1] + 3) * (i - 1)
		startXY[i] = centerY - 65
	end
	startXX[5] = PLBARS[10*5 + 1] or centerX + 40
	startXX[6] = PLBARS[10*6 + 1] or centerX + 40
	startXX[7] = PLBARS[10*7 + 1] or centerX + 40
	startXY[5] = centerY - 40
	startXY[6] = startXY[5] - PLBARS[10*5 + 4]
	startXY[7] = startXY[5] + PLBARS[10*7 + 4]
	
	-- currency
	startXX[8] = PLBARS[10*8 + 1] or centerX
	startXY[8] = centerY + PLBARS[10*8 + 4] + 20 

	-- extra
	startXX[9] = PLBARS[10*9 + 1] or startXX[5]
	startXY[9] = startXY[1]
	
	-- inv
	startXX[10] = PLBARS[10*10 + 1] or centerX - 40 - PLBARS[10*10 + 3]
	startXY[10] = centerY + PLBARS[10*10 + 4]/2

	for i = 5, barNum do
		if PLBARS[10*i + 2] then 
			startXY[i] = PLBARS[10*i + 2] + PLBARS[10*i + 4]
		else
			PLBARS[10*i + 2] = startXY[i] - PLBARS[10*i + 4]
		end
	end
	for i = 1, 4 do
		if PLBARS[10*i + 2] then
			startXY[i] = PLBARS[10*i + 2] + PLBARS[2] + PLBARS[3]
		else 
			PLBARS[10*i + 2] = startXY[i] - PLBARS[2] - PLBARS[3]
		end
	end
	
	-- create all bars
	-- createFrame(fname, x, y, w, h, RGBA, txt, overlay, r, img, parent, passive)
	-- frSub, frSubTex,   frOver, frOverTex,   frTxt,   frIco, frIcoTex
	-- createSwatch(fname, x, y, w, h, txt, txt2, RGBA)
	-- createSecureButton(bname, x, y, w, h, RGBA, img, parent, btnNum, passive)
	
	xRef[5][1], xRef[5][2], xRef[5][3], xRef[5][4] = 
	createFrame("HealthFrame", startXX[5], startXY[5], PLBARS[10*5 + 3], PLBARS[10*5 + 4], PLBARS[10*5 + 5], nil, true, ratio[5])
	
	xRef[6][1], xRef[6][2], xRef[6][3], xRef[6][4] = 
	createFrame("ManaFrame", startXX[6], startXY[6], PLBARS[10*6 + 3], PLBARS[10*6 + 4], PLBARS[10*6 + 5], nil, true, ratio[6])
	
	xRef[7][1], xRef[7][2], xRef[7][3], xRef[7][4], xRef[7][5] = 
	createFrame("CastbarFrame", startXX[7], startXY[7], PLBARS[10*7 + 3], PLBARS[10*7 + 4], PLBARS[10*7 + 5], " ", true, ratio[7])
	
	-- inv bar
	local slotWidth, slotHeight, slotY = setInvBarWHY()
	xRef[10][1], xRef[10][2] = 
	createFrame("InvFrame", startXX[10], startXY[10], PLBARS[10*10 + 3], PLBARS[10*10 + 4], PLBARS[10*10 + 5])
	-- inv bar title
	xRef[1000] = {}
		xRef[1000][1], xRef[1000][2], xRef[1000][3], xRef[1000][4], xRef[1000][5], xRef[1000][6], xRef[1000][7] = 
			createFrame("InvFrameTitle", invBar[1], PLBARS[10*10 + 4] - invBar[1], PLBARS[10*10 + 3] - 2*invBar[1], invBar[3], 
									xRGBA["pitch"], " ", nil, nil, icon[5], xRef[10][1], true)
	xRef[1000][1]:SetFrameLevel(xRef[10][1]:GetFrameLevel() + 1)
	xRef[1000][6]:SetScript("OnMouseUp", onMouseUp)
	local itemRow
	for k = 1, 18 do
		itemRow = 0
		if k > 9 then itemRow = 1 end
		if k == 10 then slotY = select(3,setInvBarWHY()) end
		xRef[1000+k] = {}
		xRef[1000+k][1], xRef[1000+k][2], xRef[1000+k][3], xRef[1000+k][4], xRef[1000+k][5], xRef[1000+k][6], xRef[1000+k][7] = 
			createFrame("InvSubFrame" .. k, invBar[1] + itemRow*(slotWidth+invBar[1]), slotY, slotWidth, slotHeight, 
									xRGBA["black"], " ", nil, nil, icon[5], xRef[10][1], true)
		xRef[1000+k][1]:SetFrameLevel(xRef[10][1]:GetFrameLevel() + 1)
		slotY = slotY - slotHeight - invBar[2]
		xRef[1000+k][6]:SetScript("OnEnter", onMouseEnter)
		xRef[1000+k][6]:SetScript("OnLeave", onMouseLeave)
	end
	toggleInvBar()
	
	-- extra bar 
	xRef[9][1], xRef[9][2] = 
	createFrame("ExtraBarFrame", startXX[9], startXY[9], PLBARS[10*9 + 3], PLBARS[10*9 + 4], PLBARS[10*9 + 5])
	xRef[9][1].text = xRef[9][1]:CreateFontString(nil,"ARTWORK")
	xRef[9][1].text:SetFontObject("GameFontHighlightLarge")
	xRef[9][1].text:SetTextColor(unpack(xRGBA["white_dim"]))
	xRef[9][1].text:SetText("EX")
	xRef[9][1].text:SetAllPoints()
	-- icon layer & texture
	_, _, _, _, _, xRef[9][5], xRef[9][3] = 
	createFrame("ExtraBarFrameIcon", 0, PLBARS[10*9 + 4]-2*extBar[1] + imgMar*2, 
							PLBARS[10*9 + 3]-2*extBar[1] + imgMar*2, PLBARS[10*9 + 4]-2*extBar[1] + imgMar*2, {1,1,1,0}, 
							nil, nil, nil, icon[5], xRef[9][1], true)
	xRef[9][4], _ = 
	createSecureButton("ExtraBarFrameButton", extBar[1], extBar[1], PLBARS[10*9 + 3]-2*extBar[1], PLBARS[10*9 + 4]-2*extBar[1],
										 {0,0,0,0}, nil, xRef[9][1], ActionButton_GetPagedID(ExtraActionBarFrame.button), true)
	xRef[9][4]:EnableMouse(true)
	xRef[9][4]:SetScript("OnMouseDown", onMouseDown)
	xRef[9][4]:SetScript("OnMouseUp", onMouseUp)
	xRef[9][4]:SetScript("OnEnter", onMouseEnter)
	xRef[9][4]:SetScript("OnLeave", onMouseLeave)
									
	-- currency bars
	xRef[8][1], xRef[8][2] = 
			createFrame("CurrFrame", startXX[8], startXY[8], PLBARS[10*8 + 3], PLBARS[10*8 + 4], PLBARS[10*8 + 5])
	for i = 1, 4 do
		xRef[80+i] = {}
		xRef[80+i][1], xRef[80+i][2], xRef[80+i][3], xRef[80+i][4], xRef[80+i][5] = 
			createFrame("CurrSubFrame" .. i, currBar[1], PLBARS[10*8 + 4] - currBar[1] - (currBar[3] + currBar[2])*(i - 1), 
									PLBARS[10*8 + 3] - 2*currBar[1], currBar[3], xRGBA["pitch"], " ", nil, nil, icon[i], xRef[8][1], true)
		xRef[80+i][1]:SetFrameLevel(xRef[8][1]:GetFrameLevel() + 1)
	end
	if PLBARS[10*8 + 6] == 0 then
		xRef[8][1]:Hide()
		for i = 1, 4 do
			xRef[80+i][1]:Hide()
		end
	end
	
	-- group swatch bars
	for i = 1, 4 do
		xRef[i][1], xRef[i][2], xRef[i][3] = 
			createSwatch("Swatch"..i, startXX[i], startXY[i], PLBARS[1], PLBARS[2], " ", " ", xRGBA["off"])
		xRef[i][4] = i
	end

	setActionBar()

	xRef[5][1]:SetScript("OnUpdate", onUpdate)
	
	print(format("|cFFffff00PlayerBars v%s loaded. Type /pbars -h to see command list.|r", ver))
	
end
end
eframe:SetScript("OnEvent", eframe.OnEvent)

local function updateSwatches()
	for i = 1, 4 do
		xRef[i][1]:SetSize(PLBARS[1], PLBARS[2])
		xRef[i][3]:SetSize(PLBARS[1], PLBARS[3])
	end
end



local cmTable = {
	["hw"] = {5,"w"},
	["hh"] = {5,"h"},
	["hc"] = {5,"c"},
	["mw"] = {6,"w"},
	["mh"] = {6,"h"},
	["mc"] = {6,"c"},
	["aw"] = {7,"w"},
	["ah"] = {7,"h"},
	["ac"] = {7,"c"},
	["iw"] = {10,"w"},
	["ih"] = {10,"h"}
}

SLASH_PLAYERBARS1 = "/pbars"
function SlashCmdList.PLAYERBARS(msg, editbox)
	local left, bottom, ind, type
	local cmd, args = strsplit(" ", strlower(msg), 2)
	
	if cmd == "" then
		if xRef[5][1]:IsVisible() then
			toggleInvBar(0)
			for i = 1, barNum do
				xRef[i][1]:Hide()
			end
			for i = 1, 4 do
				xRef[80+i][1]:Hide()
			end
			xRef[9][4]:Hide()
			xRef[9][5]:Hide()
			xRef[5][1]:SetScript("OnUpdate", nil)
		else
			toggleInvBar(1)
			for i = 1, barNum do
				xRef[i][1]:Show()
			end
			for i = 1, 4 do
				xRef[80+i][1]:Show()
			end
			xRef[9][4]:Show()
			xRef[9][5]:Show()
			xRef[5][1]:SetScript("OnUpdate", onUpdate)
		end
		return
	end
	
	if cmd == "cb" then
		if xRef[8][1]:IsVisible() then
			PLBARS[10*8 + 6] = 0
			xRef[8][1]:Hide()
			for i = 1, 4 do
				xRef[80+i][1]:Hide()
			end
		else
			PLBARS[10*8 + 6] = 1
			xRef[8][1]:Show()
			for i = 1, 4 do
				xRef[80+i][1]:Show()
			end
		end
		return
	end

	if cmd == "i" then
		if PLBARS[10*10 + 6] == 1 then
			PLBARS[10*10 + 6] = 0
		else
			PLBARS[10*10 + 6] = 1
		end
		toggleInvBar()
		return
	end

	if cmd == "x" then
		if PLBARS[10*9 + 6] == 1 then
			PLBARS[10*9 + 6] = 0
		else
			PLBARS[10*9 + 6] = 1
		end
		setExtraBar()
		return
	end
	
	if cmd == "a" then
		if PLBARS[7] == 1 then
			PLBARS[7] = 0
		else
			PLBARS[7] = 1
		end
		setActionBar()
		return
	end
	
	if cmd == "c" and args ~= "" then
		PLBARS[5] = {strsplit(" ", strlower(args), 4)}
		for i = 5, barNum do
			if i ~= 9 then xRef[i][2]:SetTexture(unpack(PLBARS[5])) end
		end
		return
	end
	
	if cmd == "f" and tonumber(args) > 0 then
		PLBARS[6] = tonumber(args)
		return
	end
	
	if cmTable[cmd] ~= nil then
		ind, type = unpack(cmTable[cmd])
		if type == "w" and tonumber(args) > 0 then
			PLBARS[10*ind + 3] = tonumber(args)
			resetFrame(xRef[ind][1])
			xRef[ind][1]:SetWidth(PLBARS[10*ind + 3])
			if xRef[ind][3] ~= nil then safeSetWidth(xRef[ind][3], ratio[ind] * PLBARS[10*ind + 3]) end
			if ind == 7 then
				xRef[ind][5]:SetWidth(PLBARS[10*ind + 3])
			end
		elseif type == "h" and tonumber(args) > 0 then
			PLBARS[10*ind + 4] = tonumber(args)
			resetFrame(xRef[ind][1])
			xRef[ind][1]:SetHeight(PLBARS[10*ind + 4])
			if xRef[ind][3] ~= nil then xRef[ind][3]:SetHeight(PLBARS[10*ind + 4]) end
			if ind == 7 then
				xRef[ind][5]:SetHeight(PLBARS[10*ind + 4])
			end
			_, PLBARS[10*ind + 2] = xRef[ind][1]:GetBoundsRect()
		elseif type == "c" and args ~= "" then
			PLBARS[10*ind + 5] = {strsplit(" ", strlower(args), 4)}
			xRef[ind][4]:SetTexture(unpack(PLBARS[10*ind + 5]))
		end
		if cmd == "iw" or cmd == "ih" then setInvBar(type) end
		return
	end
			
	if cmd == "b" then
		left, bottom = xRef[5][1]:GetBoundsRect()
		-- align widths and pos
		for i = 6, 7 do
			PLBARS[10*i + 3] = PLBARS[10*5 + 3]
			if i == 6 then 
				xRef[i][1]:SetPoint("TOPLEFT", WorldFrame, "BOTTOMLEFT", left, bottom)
			else
				xRef[i][1]:SetPoint("TOPLEFT", WorldFrame, "BOTTOMLEFT", left, bottom + PLBARS[10*5 + 4] + PLBARS[10*i + 4])
			end
			resetFrame(xRef[i][1])
			xRef[i][1]:SetWidth(PLBARS[10*i + 3])
			safeSetWidth(xRef[i][3], ratio[i] * PLBARS[10*i + 3])
			PLBARS[10*i + 1], PLBARS[10*i + 2] = xRef[i][1]:GetBoundsRect()
		end
		return
	end

	if cmd == "bgh" and args ~= "" then
		xRef[1][1]:Show()
		left, bottom = xRef[1][1]:GetBoundsRect()
		for i = 2, 4 do
			xRef[i][1]:SetPoint("TOPLEFT", WorldFrame, "BOTTOMLEFT", 
								 left + (i - 1)*(PLBARS[1] + args), 
								 bottom + PLBARS[2] + PLBARS[3])
			xRef[i][1]:Show()
			PLBARS[10 * i + 1], PLBARS[10 * i + 2] = xRef[i][1]:GetBoundsRect()
		end
		return
	end
		
	if cmd == "bgv" and args ~= "" then
		xRef[1][1]:Show()
		left, bottom = xRef[1][1]:GetBoundsRect()
		for i = 2, 4 do
			xRef[i][1]:SetPoint("TOPLEFT", WorldFrame, "BOTTOMLEFT", 
								 left, 
								 bottom + PLBARS[2] + PLBARS[3] - (i - 1)*(PLBARS[2] + PLBARS[3] + args))
			xRef[i][1]:Show()
			PLBARS[10 * i + 1], PLBARS[10 * i + 2] = xRef[i][1]:GetBoundsRect()
		end
		return
	end
			
	if cmd == "gw" and args ~= "" then
		PLBARS[1] = args
		updateSwatches()
	elseif cmd == "gh" and args ~= "" then
		PLBARS[2] = args
		updateSwatches()
	elseif cmd == "g" then
		if xRef[1][1]:IsVisible() then
			for i = 1, 4 do
				xRef[i][1]:Hide()
			end
		else
			for i = 1, 4 do
				xRef[i][1]:Show()
			end
		end
		return
	end
		
	if cmd == "-help" or cmd == "-h" or cmd == "?" or cmd == "/?" then
		for _, v in pairs(helpText) do
			print(format("%s", v))
		end
	end
	

end

